/*
 * Packet matching code.
 *
 * Copyright (C) 2008, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: bcm_nat.c,v 1.2 2009/03/30 07:50:01 Exp $
 */
#include <linux/config.h>
#include <linux/cache.h>
#include <linux/skbuff.h>
#include <linux/kmod.h>
#include <linux/vmalloc.h>
#include <linux/netdevice.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <net/route.h>
#include <net/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_nat_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_nat_lock)

#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/listhelp.h>

#include <linux/netfilter_ipv4/ip_tables.h>

#define DEBUGP(format, args...)

typedef int (*bcmNatHitHook)(struct sk_buff *skb);
extern int bcm_nat_hit_hook_func(bcmNatHitHook hook_func);

typedef int (*bcmNatBindHook)(struct ip_conntrack *ct, enum ip_conntrack_info ctinfo,
	struct sk_buff **pskb);
extern int bcm_nat_bind_hook_func(bcmNatBindHook hook_func);


extern void
manip_pkt(u_int16_t proto, struct iphdr *iph, size_t len,
	const struct ip_conntrack_manip *manip,
	enum ip_nat_manip_type maniptype,
	__u32 *nfcache);


static int
bcm_do_bindings(struct ip_conntrack *ct,
	enum ip_conntrack_info ctinfo,
	struct sk_buff **pskb)
{
	unsigned int i, j;
	unsigned int hooknum;
	enum ip_conntrack_dir dir = CTINFO2DIR(ctinfo);
	static int hn[2] = {NF_IP_PRE_ROUTING, NF_IP_POST_ROUTING};
	struct ip_nat_info *info = &ct->nat.info;

	/* Need nat lock to protect against modification, but neither
	 * conntrack (referenced) and helper (deleted with
	 * synchronize_bh()) can vanish.
	 */
	READ_LOCK(&ip_nat_lock);

	/*
	 * Do NF_IP_PRE_ROUTING and NF_IP_POST_ROUTING
	 * one in a shot.
	 */
	for (j = 0; j < 2; j++) {
		hooknum = hn[j];
		for (i = 0; i < info->num_manips; i++) {
			/* raw socket (tcpdump) may have clone of incoming
			 * skb: don't disturb it --RR
			 */
			if (skb_cloned(*pskb) && !(*pskb)->sk) {
				struct sk_buff *nskb = skb_copy(*pskb, GFP_ATOMIC);
				if (!nskb) {
					READ_UNLOCK(&ip_nat_lock);
					return NF_DROP;
				}
				kfree_skb(*pskb);
				*pskb = nskb;
			}

			if (info->manips[i].direction == dir &&
			    info->manips[i].hooknum == hooknum) {

				/*
				 * Route allocate for SRC ip chage
				 */
				if ((*pskb)->dst == NULL &&
				    info->manips[i].maniptype == IP_NAT_MANIP_SRC) {
					struct iphdr *iph = (*pskb)->nh.iph;
					struct net_device *dev = (*pskb)->dev;

					if (ip_route_input((*pskb), iph->daddr,
					    iph->saddr, iph->tos, dev)) {
						READ_UNLOCK(&ip_nat_lock);
						return NF_DROP;
					}

					/* Change skb owner */
					(*pskb)->dev = (*pskb)->dst->dev;
				}

				DEBUGP("Mangling %p: %s to %u.%u.%u.%u %u\n",
				       *pskb,
				       info->manips[i].maniptype == IP_NAT_MANIP_SRC
				       ? "SRC" : "DST",
				       NIPQUAD(info->manips[i].manip.ip),
				       htons(info->manips[i].manip.u.all));

				manip_pkt((*pskb)->nh.iph->protocol,
					(*pskb)->nh.iph,
					(*pskb)->len,
					&info->manips[i].manip,
					info->manips[i].maniptype,
					&(*pskb)->nfcache);
			}
		}
	}

	READ_UNLOCK(&ip_nat_lock);
	return NF_FAST_NAT;
}

/* 
 * Send packets to output.
 */
static inline int bcm_fast_path_output(struct sk_buff *skb)
{
	struct dst_entry *dst = skb->dst;
	struct hh_cache *hh = dst->hh;

	if (hh) {
		read_lock_bh(&hh->hh_lock);
		memcpy(skb->data - 16, hh->hh_data, 16);
		read_unlock_bh(&hh->hh_lock);
		skb_push(skb, hh->hh_len);
		return hh->hh_output(skb);
	} else if (dst->neighbour)
		return dst->neighbour->output(skb);

	kfree_skb(skb);
	return -EINVAL;
}

static int bcm_fast_path(struct sk_buff * skb)
{
	if (skb->dst == NULL) {
		struct iphdr *iph = skb->nh.iph;
		struct net_device *dev = skb->dev;

		if (ip_route_input(skb, iph->daddr, iph->saddr, iph->tos, dev)) {
			return NF_DROP;
		}
		/*  Change skb owner to output device */
		skb->dev = skb->dst->dev;
	}

	if (skb->dst) {
		if (skb->len > skb->dst->pmtu)
			return ip_fragment(skb, bcm_fast_path_output);
		else
			return bcm_fast_path_output(skb);
	}
	kfree_skb(skb);
	return NET_RX_DROP;
}

static int __init bcm_nat_init(void)
{
	int ret;
	bcm_nat_hit_hook_func(bcm_fast_path);
	bcm_nat_bind_hook_func(bcm_do_bindings);
	printk("BCM fast NAT: INIT\n");
	return 0;
}

static void __exit bcm_nat_fini(void)
{
	bcm_nat_hit_hook_func(NULL);
	bcm_nat_bind_hook_func(NULL);
}

module_init(bcm_nat_init);
module_exit(bcm_nat_fini);
MODULE_LICENSE("Proprietary");
